/*************************************************************************
Copyright (c) 2010-2011, Valentina BOEVA.

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses

>>> END OF LICENSE >>>
*************************************************************************/


// main.cpp

#include "SVfinder.h"
#include "GenomeCopyNumber.h"

using namespace std ;
int verbose = false;
int nProcesses = 1;
int degree = 3;
int minCNAlength = 1;
bool intercept;
double minMappabilityPerWindow = 0.85;
bool uniqueMatch = false;
float minExpectedGC=0.35;
float maxExpectedGC=0.55;

int main(int argc, char *argv[]) {
	//check arguments
    cout << "FREEC v5.6 (Control-FREEC v2.6) : calling copy number alterations and LOH regions using deep-sequencing data\n";
	if (argc < 3) {
		cerr << "\n\tPlease specify a config file\n\n";
		usage();
		return 0;
	}
	if (argc == 3) {
		if (
			(strcmp(argv[1], "-conf") != 0)&&
			(strcmp(argv[1], "--conf") != 0)
			) {
			usage();
			return 0;
		}
	}
	//check if config exists
    ifstream ifile(argv[2]);
    if (!ifile) {
      cerr << "\n\tCould not find your config file.. Please, check the existance of "<<argv[2] <<"\n\n";
      return -1;
    }

	//read config file
	ConfigFile cf(argv[2]);

	try {
		nProcesses = int(cf.Value("general","numberOfProcesses"));
	} catch (const char * error) {
		//Do nothing
		nProcesses = 1;
	}

	if (nProcesses>1) {
           cout << "..Unable to run " << nProcesses<<" processes\t";
           cout << "..Using only 1 process\n";
           nProcesses = 1;
	} else {
		cout << "..Using " << nProcesses<<" process(es)\n";
	}

	try {
		minCNAlength = int(cf.Value("general","minCNAlength"));
        cout << "..Minimal CNA length (in windows) was set to "<< minCNAlength<< "\n";

	} catch (const char * error) {
        cout << "..Minimal CNA length (in windows) is "<< minCNAlength<< "\n";
	}
    string sex = "";

    try {
		sex = std::string(cf.Value("general","sex"));
		if (sex.compare("XX")==0) {
            cout << "..consider the sample being female\n";
		} else if (sex.compare("XY")==0) {
            cout << "..consider the sample being male\n";
		} else {
		    cerr << "Error: \"sex\" can be either XX or XY\n";
		    exit(0);
		}
	} catch (const char * error) {
	}

    double breakPointThreshold;
	try {
		breakPointThreshold = double(cf.Value("general","breakPointThreshold"));
		if (breakPointThreshold <0) {
		    cerr << "\n\n\t!!ERROR!! (but don't be afraid :)\n\n";
            cerr << "Starting from FREEC v.4.2 we use slope of the slope threshold (instead of simply slope in FREEC v.<4.1) to define number of breakpoints in segmentation. ";
            cerr << "This method is more robust and should provide a more uniform segmentation for different chromosomes.\n";
            cerr << "\n\tWe recomend to use \"breakPointThreshold=0.8\"\n\n";
            cerr << "It should be a positive value. The higher it is, the less breakpoints you will get.\n";
            cerr << "\n\tI am sorry, but you need to change this value in your config profile.. Or your can just comment it with #, then the default values of 0.8 will be used\n";
            exit(0);
		}
	} catch (const char * error) {
		breakPointThreshold = .8;
		cout << "..breakPointThreshold set to "<<breakPointThreshold<<"\n";
	}

	try {
		degree = int(cf.Value("general","degree"));
	} catch (const char * error) {
		//Do nothing
		degree = 3;
	}
    cout << "..Polynomial degree for \"ReadCount ~ GC-content\" or \"Sample ReadCount ~ Control ReadCount\" is "<< degree<< "\n";

    int teloCentroFlanks = TELO_CENTRO_FLANCS;
    try {
		teloCentroFlanks = int(cf.Value("general","telocenromeric"));
        cout << "..telocenromeric set to "<<teloCentroFlanks<<"\n";
	} catch (const char * error) {
            //do nothing
	}


	bool contaminationAdjustment = false;
	try {
		string contam = std::string(cf.Value("general","contaminationAdjustment"));
		if (contam.compare("1")==0 || contam.compare("TRUE")==0 ||contam.compare("true")==0 ||contam.compare("True")==0)
			contaminationAdjustment = true;
	} catch (const char * error) {
        cout << "..FREEC is not going to adjust profiles for a possible contamination by normal cells\n";
	}

    float knownContamination = 0;
    try {
        knownContamination =  (float)cf.Value("general","contamination");
        if (knownContamination>0) {
            if (contaminationAdjustment == false) {
                    cerr << "..set contaminationAdjustment=TRUE if you want to use \"contamination=...\"\n";
                    exit(0);
            }
            if (knownContamination>100 ) {
                    cerr << "..contamination should not be greater than 100%\n";
                    exit(0);
            }
            if (knownContamination>1 ) {
                knownContamination /=100;
            }
            cout << "..Contamination by normal cells set to:\t" << knownContamination*100<< "%\n";
        }
        if (knownContamination<0) {
                cerr << "..contamination by normal cells should be a positive value\n";
                exit(0);
        }
        if (knownContamination==0 && contaminationAdjustment == true) {
                cerr <<  "..set contaminationAdjustment=FALSE if you don't want to use \"contamination=...\" and you think that there is no contamiantion of your tumor sample by normal cells\n";
                exit(0);
        }
	} catch (const char * error) {
		//Do nothing
	}


	string myName = "";
	string controlName = "";
	string outputDir = ".";


	try {
		outputDir = std::string(cf.Value("general","outputDir"));


        if ( access( outputDir.c_str(), 0 ) == 0 )
        {
            struct stat status;
            stat( outputDir.c_str(), &status );

            if ( status.st_mode & S_IFDIR )
            {
                cout << "..Output directory:\t" << outputDir << "\n";
            }
            else
            {
                cerr << "Error: The path you entered for 'outputDir': "<< outputDir <<" is a file. It shoud be a directory" << endl;
                exit(-1);
            }
        }
        else
        {
            cerr << "Error: Path "<<outputDir<<" doesn't exist." << endl;
            exit(-1);
        }


	} catch (const char * error) {
		//Do nothing, outputDir will be "."
	}


    string dirWithFastaSeq = "";

    try {
            dirWithFastaSeq = std::string(cf.Value("general","chrFiles"));

            if ( access( dirWithFastaSeq.c_str(), 0 ) == 0 )
            {
                struct stat status;
                stat( dirWithFastaSeq.c_str(), &status );

                if ( status.st_mode & S_IFDIR )
                {
                    cout << "..Directory with files containing chromosome sequences:\t" << dirWithFastaSeq << "\n";
                }
                else
                {
                    cerr << "Error: The path you entered for 'dirWithFastaSeq': "<< dirWithFastaSeq <<" is a file. It shoud be a directory" << endl;
                    exit(-1);
                }
            }
            else
            {
                cerr << "Error: Path "<<dirWithFastaSeq<<" doesn't exist. Comment the line with 'chrFiles' if you use a precalculated GC-content profile or set the correct path" << endl;
                exit(-1);
            }
    } catch (const char * error) {
            //Do nothing
    }

	try {
		cout << "..Sample file:\t" << std::string(cf.Value("sample","mateFile")) << "\n";
		cout << "..Sample input format:\t" << std::string(cf.Value("sample","inputFormat")) << "\n";
		myName = std::string(cf.Value("sample","mateFile"));
	} catch (const char * error) {
		try {
            cout << "..Sample file with precalculated copy numbers:\t" << std::string(cf.Value("sample","mateCopyNumberFile")) << "\n";
            myName = std::string(cf.Value("sample","mateCopyNumberFile"));
        } catch (const char * error) {
            cerr << "\n\tError: either \"mateFile\" or \"mateCopyNumberFile\" must be specified\n\n";
            return -1;
        }
	}

	try {
		cout << "..Control file with precalculated copy numbers:\t" << std::string(cf.Value("control","mateCopyNumberFile")) << "\n";
	} catch (const char * error) {
			try {
                cout << "..Control file:\t" << std::string(cf.Value("control","mateFile")) << "\n";
                cout << "..Input format for the control file:\t" << std::string(cf.Value("control","inputFormat")) << "\n";
                controlName = std::string(cf.Value("control","mateFile"));
            } catch (const char * error) {
                //Do nothing
            }
	}

// createNames:
	char rsymb = '/';
	std::vector<std::string> elems = split(myName, '\\');
	if (elems.size()>1)
		rsymb = '\\';
	myName = elems.back();
	elems.clear();
	elems = split(myName, '/' );
	if (elems.size()>1)
		rsymb = '/';
	myName = elems.back();
	if (outputDir.compare("")!=0) {
		char lsymb = outputDir.at(outputDir.length()-1);
		if (lsymb != '/' && lsymb != '\\' ) {
			outputDir = outputDir+rsymb;
		}
			myName = outputDir+myName;
	}


	elems.clear();
	if (controlName.compare("")!=0) {
		elems = split(controlName, '\\');
		controlName = elems.back();
		elems.clear();
		elems = split(controlName, '/' );
		controlName = elems.back();
		controlName = outputDir+controlName;
		elems.clear();
	}

	//try {chrLenFile
	//	GenomeDensity controlGD (cf.Value("control","mateFile"),cf.Value("control","inputFormat"),cf.Value("control","mateOrientation"),cf.Value("general","chrLenFile"));
	//} catch (const char * error) {
	//	cout << error;
	//	return -1;
	//}


	//try {
	//	GenomeDensity sampleGD (cf.Value("sample","mateFile"),cf.Value("sample","inputFormat"),cf.Value("sample","mateOrientation"),cf.Value("general","chrLenFile"));
	//} catch (const char * error) {
	//	cout << error;
	//	return -1;
	//}

    int step = NA;
    try {
        step =  (int)cf.Value("general","step");
		cout << "..Step:\t" << step<< "\n";
	} catch (const char * error) {
		//Do nothing
	}

    //READ SAMPLE DATA:

	GenomeCopyNumber sampleCopyNumber;
    float coefficientOfVariation = 0;
	if (step != NA) {
        sampleCopyNumber.setStep(step);
	}
	try {
		sampleCopyNumber.readCopyNumber(cf.Value("sample","mateCopyNumberFile"));
	} catch (const char * error) {

        try {
            cout << "..File with chromosome lengths:\t" << std::string(cf.Value("general","chrLenFile")) << "\n";
        } catch (const char * error) {
            cerr << "You need to provide a file with chromosome lengths\n";
            exit (-1);
        }
		try {
				sampleCopyNumber.readCopyNumber(cf.Value("sample","mateFile") , cf.Value("sample","inputFormat"), cf.Value("sample","mateOrientation"),cf.Value("general","chrLenFile"), (int)cf.Value("general","window"), step);
		}catch (const char * error) {
		    try {
		        coefficientOfVariation = (float)cf.Value("general","coefficientOfVariation");
		    }catch (const char * error) {
		        cerr << "Error: Either 'coefficientOfVariation' or 'window' must be specified.\nIf you use 'coefficientOfVariation' Control-FREEC will evaluate window size itself.\n";
		        cerr << "The program will continue to run with coefficientOfVariation=0.05\n";
                coefficientOfVariation=0.05;
		    }
            sampleCopyNumber.readCopyNumber(cf.Value("sample","mateFile") , cf.Value("sample","inputFormat"), cf.Value("sample","mateOrientation"),cf.Value("general","chrLenFile"), coefficientOfVariation);
            step = sampleCopyNumber.getWindowSize(); //in this case step=windowSize
		}
		sampleCopyNumber.printCopyNumber(myName+"_sample.cpn");
	}

	int wl = sampleCopyNumber.getWindowSize();
	cout << "..Window size:\t"<< wl << "\n";
	if (step == NA) {
        step= wl;
	}

	sampleCopyNumber.setSex(sex);


    //READ CONTROL DATA:


	GenomeCopyNumber controlCopyNumber;
	bool isControlIsPresent = 0;

	try {
		controlCopyNumber.readCopyNumber(cf.Value("control","mateCopyNumberFile"));
		isControlIsPresent = 1;
	} catch (const char * error) {

        try {
            string controlFile = cf.Value("control","mateFile");
            try {
                cout << "\t..Will use " << std::string(cf.Value("general","chrLenFile")) << " to calculate RC for the control sample\n";
            }  catch (const char * error) {
                cerr << "You need to provide a file with chromosome lengths to calculate Read Counts for the control sample\n";
                exit (-1);
            }
            try {
                controlCopyNumber.readCopyNumber(controlFile, cf.Value("control","inputFormat"), cf.Value("control","mateOrientation"), cf.Value("general","chrLenFile"), sampleCopyNumber.getWindowSize(), step );
                controlCopyNumber.printCopyNumber(controlName+"_control.cpn");
                isControlIsPresent = 1;
            } catch (const char * error) {
                cout << "..tried to use "<< controlFile << ", but did not succeed. \n\n\tDid you specify \"mateOrientation\" and \"inputFormat\" for the control sample?\n\t..will try GC-content normalization\n";
            }
        }
         catch (const char * error) {
             //do nothing, try the GC-content
        }

	}

    if (isControlIsPresent)
		controlCopyNumber.setSex(sex);

    //check if it is not a TARGETED resequencing experiment
    string targetBed = "";
	bool ifTargeted = false;
	try {
		targetBed = std::string(cf.Value("target","captureRegions"));
        ifTargeted = true;
        cout << "..FREEC will take into account only regions from "<<targetBed<<"\n";
        cout << "..Mappability and GC-content won't be used\n";
        minMappabilityPerWindow = 0;
        cout << "..Control-FREEC won't use minimal mappability. All windows overlaping capture regions will be considered\n";
        int minRegion = sampleCopyNumber.focusOnCapture(targetBed);
        if (teloCentroFlanks>minRegion) {
            teloCentroFlanks = minRegion;
            cout << "..telocenromeric set to "<<teloCentroFlanks<<" since it is a minimal capture region\n";
        }
        if (coefficientOfVariation) {
            cerr << "..At this point you need to profide window size, option 'window' in group of parameters [general] in your config file\n";
            exit(-1);
        }
        if  (!isControlIsPresent) {
            cerr << "..There is no control sample!!!\nYou have to use a matching control sample to get adequite results since GC-bias is not the only bias in targeted sequencing\n";
            exit(-1);
        }
        controlCopyNumber.focusOnCapture(targetBed);
	} catch (const char * error) {
	    //do nothing
	}

    //READ GC-CONTENT:
    bool wantToUseSNP = false;

    if (isControlIsPresent) {
        try {
            string SNPinfoFile = std::string(cf.Value("BAF","SNPfile"));
            if (!ifTargeted)
                cout << "..will use GC-content to normalize read count for the control sample\n";
            wantToUseSNP = 1;
        } catch (const char * error) {
            //do nothing;
        }
    }

    int forceGCwhenControlIsPresent = 0;

    if (isControlIsPresent) { //check if the user wants to use the GC-content as well
        try {
            string forceGC = std::string(cf.Value("general","forceGCcontentNormalization"));
            //exit(-1);
            if (forceGC.compare("1")==0 || forceGC.compare("TRUE")==0 ||forceGC.compare("true")==0 ||forceGC.compare("True")==0)
                forceGCwhenControlIsPresent = 1;
            if (forceGC.compare("2")==0 )
                forceGCwhenControlIsPresent = 2;
        } catch (const char * error) {
            if (!wantToUseSNP||ifTargeted)
                cout << "..FREEC is not going to use GC-content nor mappability to normalize Read Count\n";
        }
    }

    string GCprofileFile = "";

    if (ifTargeted && forceGCwhenControlIsPresent == 1) {
        cout << "..I would not advise using forceGCwhenControlIsPresent=1 for targeted resequencing since capture bias can be much stronger than GC-content bias\n";
        cout << "..Continue anyway\n";
    }

	if ((!isControlIsPresent || forceGCwhenControlIsPresent || wantToUseSNP) && !ifTargeted || forceGCwhenControlIsPresent&&ifTargeted) {//then read CG-content.
		cout << "..using GC-content to normalize the profile\n";

        try {
            minMappabilityPerWindow = double(cf.Value("general","minMappabilityPerWindow"));
            cout << "..Using the minimal mappability of: "<< minMappabilityPerWindow <<"\n";
        } catch (const char * error) {
            //Do nothing
            cout << "..Using the default minimal mappability value of "<< minMappabilityPerWindow <<"\n";
        }

        try {
            string uMatch = std::string(cf.Value("general","uniqueMatch"));
            if (uMatch.compare("1")==0 || uMatch.compare("TRUE")==0 ||uMatch.compare("true")==0 ||uMatch.compare("True")==0)
                uniqueMatch = true;
            try { //uniqueMatch can be TRUE only if a mappability file is provided
                string gemMapFile = std::string(cf.Value("general","gemMappabilityFile"));
                cout << "..Parameter uniqueMatch was set TRUE, will use "<< gemMapFile <<" for mappability information\n";
            } catch (const char * error) {
                uniqueMatch = false;
                cout << "..Parameter uniqueMatch was set FALSE, to set it TRUE provide a GEM mappability file\n";
            }

        } catch (const char * error) {

            //Do nothing
            uniqueMatch = false;
            cout << "..Parameter uniqueMatch was set FALSE\n";
        }

		try {
			//check, maybe a file with CG-content already exists?
            GCprofileFile = std::string(cf.Value("general","GCcontentProfile"));
			int stepGC = sampleCopyNumber.readCGprofile(GCprofileFile);
			if (step!=stepGC) {
			    cerr << "Error: Uncorrect window size in the GC-content profile. Will need to recalculate it. You must provide a path to chromosome files, option \"chrFiles\"\n";
                throw "Error: Uncorrect window size in the GC-content profile.\n";
			}
           // cout << "..file "<< GCprofileFile << " is read\n";
		} catch (const char * error) { // create the profile from genomic sequences
		    if (dirWithFastaSeq=="") {
                cerr << "..since GCcontentProfile is not provided, FREEC needs \"chrFiles\" to calculate it..";
                cerr << "Please provide the directory with fasta sequences - chrFiles!!!";
                exit(0);
		    }
			sampleCopyNumber.fillCGprofile(dirWithFastaSeq);
			GCprofileFile = outputDir+"GC_profile.cnp";
			sampleCopyNumber.printCGprofile(GCprofileFile);
		}
		try { //read mappability file
            sampleCopyNumber.readGemMappabilityFile(cf.Value("general","gemMappabilityFile"));
            //rewrite GC-profile with mappability as the last (5th) colomn
            GCprofileFile = outputDir+"GC_profile.cnp";
            sampleCopyNumber.printCGprofile(GCprofileFile);
            cout << "..Mappability track from "<<cf.Value("general","gemMappabilityFile")<<" has been added to "<< GCprofileFile <<"\n";
        } catch (const char * error) {
            //
        }
	}

	if ((isControlIsPresent && (forceGCwhenControlIsPresent==1 || wantToUseSNP==true) && !ifTargeted || ifTargeted&&(forceGCwhenControlIsPresent==1)) ) {//then read CG-content and associate it with the control data.
		cout << "..using GC-content to normalize the control profile\n";

		try {
			//the file with CG-content already exists
			controlCopyNumber.readCGprofile(GCprofileFile);
			if (ifTargeted) {
                controlCopyNumber.focusOnCapture(targetBed); // to mask again averything which is not in the capture
			}
		} catch (const char * error) { // SHOULD NOT GET THERE
			cerr << "..need "<< GCprofileFile << " to use the GC-content to normalize the control profile\n";
			exit(-1);
		}
	}

    //NORMALIZE READ COUNT:

	if (isControlIsPresent) {
        try {
            intercept = (bool)int(cf.Value("general","intercept"));
        } catch (const char * error) {
            intercept = 0;
        }
		//check if window size is the same for the Control and Sample
		if (sampleCopyNumber.getWindowSize() != controlCopyNumber.getWindowSize()) {
			cerr << "\nError: the window length is different for sample and control data\n\tPlease check parameters!\n\n";
			return -1;
		}
		try {
			//at this moment we need to know genome ploidy
			sampleCopyNumber.setPloidy(round(float(cf.Value("general","ploidy"))));
		} catch (const char * error) {
			cerr << "at this moment we need to know the genome ploidy.. unable to proceed..";
			return -1;
		}
		if (!forceGCwhenControlIsPresent && !wantToUseSNP || ifTargeted&&forceGCwhenControlIsPresent!=1) {
            sampleCopyNumber.calculateRatio(controlCopyNumber, degree,intercept);
		} else {
		    if (forceGCwhenControlIsPresent==1) { //normalize first Sample and Control, and then calculate the ratio

                try {
                    intercept = (bool)int(cf.Value("general","intercept"));
                } catch (const char * error) {
                    intercept = 1;
                }
                try {
                //at this moment we need to know genome ploidy
                    sampleCopyNumber.setPloidy(round(float(cf.Value("general","ploidy"))));
                    controlCopyNumber.setPloidy(2); // normal genome has ploidy=2!!!
                    cout << "..Set ploidy for the control genome equal to "<< 2 << "\n";
                } catch (const char * error) {
                    cerr << "\nError: at this moment we need to know the genome ploidy.. unable to proceed..\n\n";
                    return -1;
                }
                try {
                    minExpectedGC = float(cf.Value("general","minExpectedGC"));
                } catch (const char * error) {
                    //do nothing
                }
                cout << "..minimal expected GC-content (general parameter \"minExpectedGC\") was set to "<< minExpectedGC<<"\n";
                try {
                    maxExpectedGC = float(cf.Value("general","maxExpectedGC"));
                } catch (const char * error) {
                    //do nothing
                }
                cout << "..maximal expected GC-content (general parameter \"maxExpectedGC\") was set to "<< maxExpectedGC <<"\n";
                sampleCopyNumber.calculateRatioUsingCG(degree, intercept,minExpectedGC,maxExpectedGC);
                controlCopyNumber.calculateRatioUsingCG(degree, intercept,minExpectedGC,maxExpectedGC);

                sampleCopyNumber.calculateRatioUsingCG(controlCopyNumber);
            } else if (forceGCwhenControlIsPresent==2) {  //calculate the ratio , normalize for GC

                sampleCopyNumber.calculateRatio(controlCopyNumber, degree,intercept);
                try {
                    minExpectedGC = float(cf.Value("general","minExpectedGC"));
                } catch (const char * error) {
                    //do nothing
                }
                cout << "..minimal expected GC-content (general parameter \"minExpectedGC\") was set to "<< minExpectedGC<<"\n";
                try {
                    maxExpectedGC = float(cf.Value("general","maxExpectedGC"));
                } catch (const char * error) {
                    //do nothing
                }
                cout << "..maximal expected GC-content (general parameter \"maxExpectedGC\") was set to "<< maxExpectedGC <<"\n";
                sampleCopyNumber.recalculateRatioUsingCG(8, intercept,minExpectedGC,maxExpectedGC); //try higher values of polynomial's degree

		    }
		}
        if(wantToUseSNP && forceGCwhenControlIsPresent!=1 && !ifTargeted) {
            intercept = 1;

            controlCopyNumber.setPloidy(2); //normal genome has ploidy=2
            cout << "..Set ploidy for the control genome equal to "<< 2 << "\n";

            try {
                    minExpectedGC = float(cf.Value("general","minExpectedGC"));
            } catch (const char * error) {
                    //do nothing
            }
            cout << "..minimal expected GC-content (general parameter \"minExpectedGC\") was set to "<< minExpectedGC<<"\n";
            try {
                maxExpectedGC = float(cf.Value("general","maxExpectedGC"));
            } catch (const char * error) {
                    //do nothing
            }
            cout << "..maximal expected GC-content (general parameter \"maxExpectedGC\") was set to "<< maxExpectedGC <<"\n";

            if (forceGCwhenControlIsPresent==0) {
                sampleCopyNumber.calculateRatioUsingCG(degree, intercept,minExpectedGC,maxExpectedGC);
		    }
            controlCopyNumber.calculateRatioUsingCG(degree, intercept,minExpectedGC,maxExpectedGC);

        }
        if (ifTargeted && wantToUseSNP && forceGCwhenControlIsPresent!=1) {
            cout << "Control-FREEC will assume that there is not gains and losses in the target regions in the control genome\n";
            controlCopyNumber.setPloidy(2); //normal genome has ploidy=2
            cout << "..Set ploidy for the control genome equal to "<< 2 << "\n";
            controlCopyNumber.setAllNormal();
        }

	} else {
        try {
            intercept = (bool)int(cf.Value("general","intercept"));
        } catch (const char * error) {
            intercept = 1;
        }
		try {
			//at this moment we need to know genome ploidy
			sampleCopyNumber.setPloidy(round(float(cf.Value("general","ploidy"))));
		} catch (const char * error) {
			cerr << "\nError: at this moment we need to know the genome ploidy.. unable to proceed..\n\n";
			return -1;
		}
        try {
            minExpectedGC = float(cf.Value("general","minExpectedGC"));
        } catch (const char * error) {
            //do nothing
        }
        cout << "..minimal expected GC-content (general parameter \"minExpectedGC\") was set to "<< minExpectedGC<<"\n";
        try {
            maxExpectedGC = float(cf.Value("general","maxExpectedGC"));
        } catch (const char * error) {
            //do nothing
        }
        cout << "..maximal expected GC-content (general parameter \"maxExpectedGC\") was set to "<< maxExpectedGC <<"\n";
		sampleCopyNumber.calculateRatioUsingCG(degree, intercept,minExpectedGC,maxExpectedGC);
	}
	cout << "..Copy number profile normalization -> done\n";

	//segmentation:

	int breakPointType;
	try {
		breakPointType = int(cf.Value("general","breakPointType"));
        cout << "..break-point type set to "<<breakPointType<<"\n";
	} catch (const char * error) {
		breakPointType = NORMALLEVEL;
		cout << "..break-point type set to "<<breakPointType<<"\n";
	}

    if (knownContamination>0) {
        cout << "..Recalculating copy number profiles using known value of contamination by normal cells:\n";
        cout << ".."<< knownContamination*100 <<"%\n";
        sampleCopyNumber.recalculateRatio(knownContamination);
        sampleCopyNumber.setNormalContamination(knownContamination);
    }


	sampleCopyNumber.calculateBreakpoints(breakPointThreshold,breakPointType);

	if (wantToUseSNP)
        controlCopyNumber.calculateBreakpoints(breakPointThreshold,breakPointType);



	cout << "..calculate median values\n";
	sampleCopyNumber.calculateCopyNumberMedians(minCNAlength);
	//sampleCopyNumber.calculatePloidy(minCNAlength);
	sampleCopyNumber.recalcFlanks(teloCentroFlanks, 3);
	//sampleCopyNumber.deleteFlanks(TELO_CENTRO_FLANCS);
	cout << "..annotate copy numbers\n";
	sampleCopyNumber.calculateCopyNumberProbs_and_genomeLength(breakPointType);

	if (contaminationAdjustment && knownContamination==0) {
	    cout <<"..Evaluating possible contamination..\n";
		float contamValue=sampleCopyNumber.evaluateContamination();
		cout << "..Identified contamination by normal cells: " << contamValue*100 << "%\n";
		if (contamValue>0) {
		    cout << "..Recalculating copy number profiles..\n";
            sampleCopyNumber.recalculateRatio(contamValue);
            knownContamination = contamValue;
            sampleCopyNumber.setNormalContamination(knownContamination);
            cout << "..Recalculate breakpoints\n";
            sampleCopyNumber.calculateBreakpoints(breakPointThreshold,breakPointType);
            cout << "..Recalculate median values\n";
            sampleCopyNumber.calculateCopyNumberMedians(minCNAlength);
            //sampleCopyNumber.deleteFlanks(TELO_CENTRO_FLANCS);
            sampleCopyNumber.recalcFlanks(teloCentroFlanks, 3);
            cout << "..Reannotate copy numbers\n";
            sampleCopyNumber.calculateCopyNumberProbs_and_genomeLength(breakPointType);
            //sampleCopyNumber.printRatio(myName+"_ratio.txt");
		}
	}
    try {
        string SNPinfoFile = std::string(cf.Value("BAF","SNPfile"));
        cout << "..will use SNP positions from "<< SNPinfoFile << " to calculate BAF profiles\n";
        string inputFormat = cf.Value("sample","inputFormat");
        SNPinGenome snpingenome(SNPinfoFile);
        int minimalTotalLetterCountPerPosition = 0;

        try {
            minimalTotalLetterCountPerPosition =  round(float(cf.Value("BAF","minimalCoveragePerPosition")));
            cout << "..will a threshold of "<< minimalTotalLetterCountPerPosition <<" read(s) per SNP position to calculate beta allel frequency (BAF) values\n";
        } catch (const char * error) {
            //Do nothing
        }


        snpingenome.assignValues(cf.Value("sample","mateFile"), inputFormat, minimalTotalLetterCountPerPosition);
        cout << "..Adding BAF info to the Sample dataset\n";
        sampleCopyNumber.addBAFinfo(snpingenome);
        cout << "..Recalculate breakpoints using BAF profiles\n";
        breakPointThreshold = 0.8;
        sampleCopyNumber.calculateBAFBreakpoints(breakPointThreshold,breakPointType);
        cout << "..Recalculate median values\n";
        sampleCopyNumber.calculateCopyNumberMedians(minCNAlength);
        //sampleCopyNumber.recalcFlanks(TELO_CENTRO_FLANCS, 3); will not do it here..
        cout << "..Reannotate copy numbers\n";
        sampleCopyNumber.calculateCopyNumberProbs_and_genomeLength(breakPointType);
	    sampleCopyNumber.printBAF(myName+"_BAF.txt",snpingenome);

        //the same for the control sample:
        if (isControlIsPresent) {

            inputFormat = (string)cf.Value("control","inputFormat");
            SNPinGenome snpingenomeControl(SNPinfoFile);
            snpingenomeControl.assignValues(cf.Value("control","mateFile"), inputFormat, minimalTotalLetterCountPerPosition);
            cout << "..Adding BAF info to the Control dataset\n";
            controlCopyNumber.addBAFinfo(snpingenomeControl);
            cout << "..Recalculate breakpoints using BAF profiles\n";
            controlCopyNumber.calculateBAFBreakpoints(breakPointThreshold,breakPointType);
            cout << "..Recalculate median values\n";
            controlCopyNumber.calculateCopyNumberMedians(minCNAlength);
            //sampleCopyNumber.recalcFlanks(TELO_CENTRO_FLANCS, 3); will not do it here..
            cout << "..Reannotate copy numbers\n";
            controlCopyNumber.calculateCopyNumberProbs_and_genomeLength(breakPointType);

            sampleCopyNumber.calculateSomaticCNVs(controlCopyNumber.getCNVs(),controlCopyNumber.getPloidy());

            controlCopyNumber.printBAF(myName+"_normal_BAF.txt",snpingenomeControl);
            controlCopyNumber.printRatio(myName+"_normal_ratio.txt");
            controlCopyNumber.printCNVs(myName+"_normal_CNVs");
        }

    } catch (const char * error) {
            //Do nothing
    }


//	sampleCopyNumber.printRatio("chr1", "ratio1.txt");
//	sampleCopyNumber.printRatio("chr5", "ratio5.txt");
	sampleCopyNumber.printRatio(myName+"_ratio.txt");
	sampleCopyNumber.printCNVs(myName+"_CNVs");

	return 0;
}
